// Glaze Library
// For the license information refer to glaze.hpp

// Deprecated: glz::tuple now lives in "glaze/core/tuple.hpp". This header remains so that
// existing includes of <glaze/tuplet/tuple.hpp> keep working, and will be removed in a
// future major release.

#pragma once

#include "glaze/core/tuple.hpp"
